int ku_c=-1;

#define MAX_KU_PROBEN 2048
#define MAX_DIV_STRING 65536

/* eigenes Alphabet */
unsigned char alpha[256];
int anzalpha;

/* Zuordnung String zu Liste */

int zu[MAX_DIV_STRING];


/* Listen mit Elementen aus dem Pattern, immer row,col-Werte nacheinander, 255 ist das Ende Zeichen */
/* maximal 16384 verschiedene Proben an jeweils maximal 2048 Positionen */
unsigned char ku_liste[16384][MAX_KU_PROBEN];
int lastkuliste;

/* Listen initialisieren */
void init_ku(int c) {
	int i1,i2;
	for(i1=0; i1<256; i1++)
		alpha[i1]=255;
	anzalpha=0;
	for(i1=0; i1<65536; i1++)
		zu[i1]=-1;
	for(i1=0; i1<16384; i1++) {
		for(i2=0; i2<MAX_KU_PROBEN; i2++) {
			ku_liste[i1][i2]=255;
		}
	}
	lastkuliste=0;
	ku_c=c;
}

/* Umwandlung eines Zeichens in die interne Darstellung als int */
int char2int(unsigned char c1) {
	if(alpha[c1] == 255) {
		/* new alphabet entry */
		alpha[c1]=anzalpha;
		anzalpha++;
		if(ku_c >= 0 && anzalpha > ku_c) {
			printf("ERROR: Das wirkliche Alphabet ist groesser als das Uebergebene!\n\n");
			exit(-1);
		}
	}
	return((int)alpha[c1]);
}

/* positive Potenzen ret=a^b */
unsigned pot(unsigned a, unsigned b) {
	unsigned u1;
	unsigned ret;

	if(b == 0)
		return(1);
	else
		ret=a;

	for(u1=1; u1<b; u1++) {
		ret*=a;
	}

	return(ret);
}

/* Abgleich von Muster und Text */
/* Uebergeben werden die Positionen der Proben in */
/* Pattern und Text */

int check_ku_trivial(pattern_row, pattern_col, text_row, text_col) {
	int row,col;
	int i,j;
	int found;

	/* Umrechnung auf (0,0) des Pattern */
	row=text_row-pattern_row;
	col=text_col-pattern_col;

#ifdef DEBUG
	printf("Untersuche: (%i,%i)\n",row,col);
#endif

	/* trivialer Abgleich */
	found=1;
	for(i=0; i<m; i++) {
		for(j=0; j<m; j++) {
			if(pattern[i][j] != text[i+row][j+col]) {
				i=m;
				j=m;
				found=0;
			}
		}
	}
	return(found);
}

/* Quadratische Suche */
void ku_quadratic_search_pre(void) {
	int i,j,qi,qr,qc,i1,ku_q,ku_sq;
	unsigned string;
	int schranke;
		
	/* Bestimmung von Q */
	ku_q=(int)ceil(log(m*m)/log(ku_c));
	if(ku_q<1)
		ku_q=1;
	ku_sq=(int)ceil(sqrt(ku_q));

#ifdef DEBUG
	printf("m=%i\nn=%i\nq=%in",m,n,ku_q);
#endif

	if(pot(ku_c,ku_q) >= MAX_DIV_STRING) {
		printf("ERROR: Zuviele verschiedene Probenketten mglich!\n\n");
		exit(-1);
	}

	/* =========================== */
	/* Vorverarbeitung des Pattern */
  /* =========================== */	
	schranke=m-ku_sq;
	for(i=0; i<=schranke; i++) {
		for(j=0; j<=schranke; j++) {
			/* (i,j) ist die Position der Probe im Pattern */
			string=0;
			for(qi=0; qi<ku_q; qi++) {
				qr=qi/ku_sq;
				qc=qi%ku_sq;
				/* (qr,qc) ist die aktuelle Position in der Probe */
				string*=ku_c;
				string+=char2int(pattern[i+qr][j+qc]);
			}
#ifdef DEBUG
			printf("\nString at (%i,%i) = %u\n",i,j,string);
#endif
			if(zu[string] == -1) {
				/* neue Liste angelegt */
				zu[string]=lastkuliste;
				ku_liste[lastkuliste][0]=i;
				ku_liste[lastkuliste][1]=j;
				lastkuliste++;
			} else {
				/* neues Element in alter Liste */
				i1=2;
				while(ku_liste[zu[string]][i1] != 255 && i1 < MAX_KU_PROBEN) 
					i1++;
				if(i1 == MAX_KU_PROBEN) {
					printf("ERROR: Zu viele gleiche Proben im Pattern!\n");
					exit(-1);
				}
				ku_liste[zu[string]][i1]=i;
				ku_liste[zu[string]][i1+1]=j;
			}
		}
	}
#ifdef DEBUG
	printf("AnzAlpha=%u\nAnzahl Liste=%i\n",anzalpha,lastkuliste);
#endif
}

/* Lineare Suche */
void ku_linear_search_pre(void) {
  int i,j,qi/*,qr,qc*/,i1,ku_q,ku_h;
	int oldleft;
  unsigned string;

  /* Bestimmung von Q */
  ku_q=(int)ceil(log(m*m)/log(ku_c));
	if(ku_q<1)
		ku_q=1;

	/* Bestimmung von h */
	ku_h=floor(m/ku_q);

#ifdef DEBUG
  printf("m=%i\nn=%i\nq=%i\nh=%i\n",m,n,ku_q,ku_h);
#endif

	if(pot(ku_c,ku_q) >= MAX_DIV_STRING) {
		printf("ERROR: Zuviele verschiedene Probenketten mglich!\n\n");
		exit(-1);
	}

  /* =========================== */
  /* Vorverarbeitung des Pattern */
  /* =========================== */
  for(i=0; i<m; i++) {
		for(j=0; j<ku_h; j++) {
			string=0;
			for(qi=0; qi<=(ku_q-1)*ku_h; qi+=ku_h) {
				string*=ku_c;
				string+=char2int(pattern[i][j+qi]);
			}	

#ifdef DEBUG
      printf("\nString at (%i,%i) = %u\n",i,j,string);
#endif
      if(zu[string] == -1) {
        /* neue Liste angelegt */
        zu[string]=lastkuliste;
        ku_liste[lastkuliste][0]=i;
        ku_liste[lastkuliste][1]=j;
        lastkuliste++;
      } else {
        /* neues Element in alter Liste */
        i1=2;
        while(ku_liste[zu[string]][i1] != 255 && i1 < MAX_KU_PROBEN)
          i1++;
        if(i1 == MAX_KU_PROBEN) {
          printf("ERROR: Zu viele gleiche Proben im Pattern!\n");
          exit(-1);
        }
        ku_liste[zu[string]][i1]=i;
        ku_liste[zu[string]][i1+1]=j;
      }
    }
  }
#ifdef DEBUG
  printf("AnzAlpha=%u\nAnzahl Liste=%i\n",anzalpha,lastkuliste);
#endif
}



/* Quadratische Suche */
void ku_quadratic_search(void) {
  int i,j,qi,qr,qc,i1,ti,tj,ku_q,ku_sq;
  unsigned string;
  int schranke,miss;

  /* Bestimmung von Q */
	/* koennte man auch global bestimmen, da schon in der */
  /* Vorverarbeitung benutzt, spart aber nicht viel Zeit */
  ku_q=(int)ceil(log(m*m)/log(ku_c));
	if(ku_q<1)
		ku_q=1;
  ku_sq=(int)ceil(sqrt(ku_q));

	/* ========================== */
	/* Durchsuchen vom Text, Test */
	/* ========================== */

	schranke=(int)floor((n-ku_sq+1)/(m-ku_sq+1));

	for(i=1; i<=schranke; i++) {
		for(j=1; j<=schranke; j++) {
			ti=i*(m-ku_sq+1)-1;
			tj=j*(m-ku_sq+1)-1;
			
			/* (ti,tj) ist die Koordinate der Probe im Text */

			string=0;
			miss=0;
			for(qi=0; qi<ku_q; qi++) {
				qr=qi/ku_sq;
				qc=qi%ku_sq;
				/* (qr,qc) ist die aktuelle Position in der Probe */

				if(alpha[text[ti+qr][tj+qc]] == 255) {
					/* das Zeichen existiert nicht im Pattern, kann keine gueltige Probenkette geben! */
					miss=1;
				} else {
					string*=ku_c;
					string+=alpha[text[ti+qr][tj+qc]];
				}
			}
#ifdef DEBUG
			printf("Probe im Text (%i,%i): %u - %i\n",ti,tj,string,miss);
#endif

			if(miss == 0) {

				i1=0;
				while(ku_liste[zu[string]][i1] != 255 && i1 < MAX_KU_PROBEN) {
#ifdef DEBUG
					printf("Position im Pattern: (%i,%i) mit (%i,%i) im Text\n",ku_liste[zu[string]][i1],ku_liste[zu[string]][i1+1],ti,tj);
#endif
					/* Ueberpruefung */
					if(check_ku_trivial(ku_liste[zu[string]][i1],ku_liste[zu[string]][i1+1],ti,tj) == 1) {
#ifdef DEBUG
						printf(" ***** Fundstelle: (%i,%i)\n",ti-ku_liste[zu[string]][i1],tj-ku_liste[zu[string]][i1+1]);
#endif
#ifdef ERGOUT
						printf("Gefunden bei (%i,%i)\n",ti-ku_liste[zu[string]][i1],tj-ku_liste[zu[string]][i1+1]);
#endif
					}
					i1+=2;
				}

			}
		}
	}
}

/* Lineare Suche */
void ku_linear_search(void) {
  int i,j,qi,qr,qc,i1,ti,tj,ku_q,ku_h;
	int stellenwech, letztestelle;
  unsigned string;
  int miss;

  /* Bestimmung von Q und h */
  /* koennte man auch global bestimmen, da schon in der */
  /* Vorverarbeitung benutzt, spart aber nicht viel Zeit */
  ku_q=(int)ceil(log(m*m)/log(ku_c));
	if(ku_q<1)
		ku_q=1;

	ku_h=floor(m/ku_q);

  /* ========================== */
  /* Durchsuchen vom Text, Test */
  /* ========================== */

	stellenwech=(pot(ku_c,(ku_q-1)));
	letztestelle=(ku_q-1)*ku_h;

  for(i=1; i<=floor(n/m); i++) {
    for(j=1; j<=floor((n-m)/ku_h)+1; j++) {
      ti=i*m-1;
      tj=j*ku_h-1;

      /* (ti,tj) ist die Koordinate der Probe im Text */

      //string=0;
      miss=0;
			if(j == 1) {
				string=0;
      	for(qi=0; qi<=(ku_q-1)*ku_h; qi+=ku_h) {
					if(alpha[text[ti][tj+qi]] == 255) {
						miss=1;
					} else {
  		      string*=ku_c;
    		    string+=alpha[text[ti][tj+qi]];
					}
				}
      } else {
				if(alpha[text[ti][tj+letztestelle]] == 255) {
					miss=1;
				} else {
					string%=stellenwech;
					string*=ku_c;
					string+=alpha[text[ti][tj+letztestelle]];
				}
			}
#ifdef DEBUG
      printf("Probe im Text (%i,%i): %u - %i\n",ti,tj,string,miss);
#endif

      if(miss == 0) {

        i1=0;
        while(ku_liste[zu[string]][i1] != 255 && i1 < MAX_KU_PROBEN) {
#ifdef DEBUG
          printf("Position im Pattern: (%i,%i) mit (%i,%i) im Text\n",ku_liste[zu[string]][i1],ku_liste[zu[string]][i1+1],ti,tj);
#endif
          /* Ueberpruefung */
          if(check_ku_trivial(ku_liste[zu[string]][i1],ku_liste[zu[string]][i1+1],ti,tj) == 1) {
#ifdef DEBUG
            printf(" ***** Fundstelle: (%i,%i)\n",ti-ku_liste[zu[string]][i1],tj-ku_liste[zu[string]][i1+1]);
#endif
#ifdef ERGOUT
            printf("Gefunden bei (%i,%i)\n",ti-ku_liste[zu[string]][i1],tj-ku_liste[zu[string]][i1+1]);
#endif
          }
          i1+=2;
        }
      }
    }
  }
}

